appversion = "0.2"

-- Terminate any WLAN connection that hasn't been closed
Wlan.term()

-- Set buffers and colors for the graphicsPrint function
white = Color.new(255, 255, 255)
offscreen = Image.createEmpty(480, 272)
offscreen:clear(Color.new(0, 0, 0))
y = 0
x = 0

-- 
--  Print text on the screen, console-format - taken from WLAN example, with some
--  modifications to allow text to overflow the screen.
-- 
function graphicsPrint(text)
	if y > 272 then
		offscreen:clear(Color.new(0, 0, 0))
		y = 0
	end
	for i = 1, string.len(text) do
		char = string.sub(text, i, i)
		if char == "\n" then
			y = y + 8
			x = 0
		elseif char ~= "\r" then
			offscreen:print(x, y, char, white)
			x = x + 8
			if x > 480 then
				y = y + 8
				x = 0
			end
		end
	end
	screen:blit(0, 0, offscreen)
	screen.waitVblankStart()
	screen.flip()
end

settingsfile = io.open("config.txt","r")
wlan_id = tonumber(settingsfile:read("*l"))
chunksize = tonumber(settingsfile:read("*l"))
settingsfile:close()

mime = { png="image/png", gif="image/gif", jpg="image/jpeg", txt="text/plain", htm="text/html", html="text/html" }

-- init WLAN and choose connection config
Wlan.init()
graphicsPrint("PSP-HTTPD beta " .. appversion .. "\n")
graphicsPrint("by Elxx - http://elxx.net\n")
graphicsPrint("Uses portions of code from Steve Kemp\n\n")
configs = Wlan.getConnectionConfigs()
graphicsPrint("Available WLAN connections:\n")
for key, value in configs do
	graphicsPrint(key .. ": " .. value .. "\n")
end
graphicsPrint("\n")
graphicsPrint("Connection #" .. wlan_id .. " will be used.\n")
graphicsPrint("Data chunk size for file transfer is " .. chunksize .. " bytes.\n")
graphicsPrint("You can change these settings in config.txt.\n\n")
Wlan.useConnectionConfig(wlan_id)

function urldecode(str) -- Decode special chars in URL's such as %20
	str = string.gsub (str, "+", " ") 
	str = string.gsub (str, "%%(%x%x)", function(h) return string.char(tonumber(h,16)) end) 
	str = string.gsub (str, "\r\n", "\n") 
	return str 
end

function fileSize(file)
	local current = file:seek()      -- get current position
	local size = file:seek("end")    -- get file size
	file:seek("set", current)        -- restore position
	return size
end

function friendlySize(size)
	if size >= 1048576 then				-- 1,048,576b = 1 MB
		return (size / 1048576) .. " MB"
	elseif size >= 1024 then				-- 1024b = 1 kB
		return (size / 1024) .. " kB"
	else 
		return size .. " bytes"
	end
end	

function string.endsWith(String,End)
	return End=='' or string.sub(String,-string.len(End))==End
end

function start_server(port)
	running = 1;

	--  Create a server socket to listen on the port
	serverSocket = Socket.createServerSocket( port )

	--   Print some status messages.
	graphicsPrint("Server socket listening on port " .. port .. "\n");
	graphicsPrint("If the green WLAN status light on your PSP\n")
	graphicsPrint("turns off, the server will not function.\n\n")

	--  Loop accepting requests.
	while ( running == 1 ) do
		status = processConnection( serverSocket )
		if status == "stop" then running = 0 end
	end

	-- Server stopped, close the socket.
	serverSocket:close()
end

function processConnection( serverSocket ) 
    --  Accept a new connection

	found   = 0;  -- Found the end of the HTTP headers?
	chunk   = 0;  -- Count of data read from client socket
	size    = 0;  -- Total size of incoming request.
	code    = 0;  -- Status code we send to the client
	request = ""; -- Request body read from client.

	--  Read in a response from the client, terminating at the first
	-- '\r\n\r\n' line - this is the end of the HTTP header request.
	while ( found == 0 ) do
		client = serverSocket:accept()
		if Controls.read():start() then break end
		if client then
			graphicsPrint("Client found.\n")
			while (found == 0) do
				data = client:recv()
				size    = size + string.len(data)
				request = request .. data
				position,len = string.find( request, '\r\n\r\n' )
				if ( position ~= nil )  then
					found = 1
				end
				if Controls.read():start() then return "stop" end
			end
		end
    end

    -- If we got this far without a client, that means Start was pressed.
    -- This doesn't work very cleanly yet...

	if not client then 
		running = 0
	end

	--  We now have a complete HTTP request of 'size' bytes long stored in 'request'.
    
	-- Find the requested path.
	_, _, method, path, major, minor  = string.find(request, "([A-Z]+) (.+) HTTP/(%d).(%d)")

	-- Decode the requested path.
	path = urldecode( path )
	
	-- Show the requested path on the PSP.
	graphicsPrint( path .. "\n")

	-- Find the user agent and referrer in case we need to log this someday.
	-- Virtual hosts are not yet supported.

	_, _, agent   = string.find(request, "Agent: ([^\r\n]+)")
	_, _, referer = string.find(request, "Referer: ([^\r\n]+)")
	-- _, _, host    = string.find(request, "Host: ([^:\r\n]+)")

	-- If there was no "Host:" header then use "default".
	if ( host == nil ) then
		host = "default"
	end

	-- If the request was for '/finish', then stop the server.
	if ( path == "/finish" ) then
		running = 0;
		client:send( "HTTP/1.0 200 OK\r\n" )
		client:send( "Server: lua-httpd " .. appversion .. "\r\n" )
		client:send( "Content-type: text/html\r\n" )
		client:send( "Connection: close\r\n\r\n" )
		client:send( "<html><head><title>Server Terminated</title></head>" )
		client:send( "<body><p>PSP-httpd has been terminated.</p></body></html>")
	else
		-- Otherwise, handle the request.
		size, code = handleRequest( path, client )
	end

	-- if ( agent == nil )   then agent   = "-" end
	-- if ( referer == nil ) then referer = "-" end
	-- if ( code == nill )   then code    = 0 ; end

	if ( running == 0 ) then
		graphicsPrint( "Server terminated." )
	end

    --
    --  Close the client connection.
    --
    client:close();
end

function handleDirectory( client, file, path )			

	client:send("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n")
	client:send("<html><head><title>PSP Web Server</title></head><body><h1>Browsing Directory: " .. file .. "</h1>\r\n")

	files = System.listDirectory(file)
	for index, file in files do
		if file.directory and file.name ~= "." then
			client:send("<p><a href=\"" .. file.name .. "/" .. "\">" .. file.name .. "</a></p>")
		end
	end
	for index, file in files do
		if not file.directory then
			client:send("<p><a href=\"" .. file.name .. "\">" .. file.name .. "</a> (" .. friendlySize(file.size) .. ")</p>")
		end
	end

	client:send("</body>")
end

function fileExists(file)
	t, msg = io.open(file, "r")
     if not t then return false end
	return true
end

function handleRequest( path, client )
	-- Local file
	file = path

	--  File must be on the Memory Stick.
	file = "ms0:" .. file

	-- If it is a directory, handle it.
	--
	if ( string.endsWith( file, "/" ) ) then
		if (fileExists( file .. "index.htm" )) then 
			file = file .. "index.htm"
		elseif (fileExists( file .. "index.html" )) then
			file = file .. "index.html"
		else
			size = handleDirectory( client, file, path )
			client:close()
			return size, "200"
		end
	end

	-- Open the file and return an error if it fails.
	if ( fileExists( file ) == false ) then 
		client:close()
		return size, "404"
	end;

	-- Find the suffix to get the mime.type.
	_, _, ext  = string.find( file, "\.([^\.]+)$" )
	ext = string.lower(ext)
	if ( ext == nil ) then
		ext = ""
	end
    
	type = mime[ext];
	if ( type == nil ) then type = 'text/plain' end

	-- Send out the header.
	client:send( "HTTP/1.0 200 OK\r\n" )
	client:send( "Server: lua-httpd " .. appversion .. "\r\n" )
	client:send( "Content-type: " .. type  .. "\r\n" )
	client:send( "Connection: close\r\n\r\n" )

	-- Read the file, and then serve it.
	-- I must warn you now that this is still buggy, and there's a chance of
	-- larger binary files being corrupted. I'm working on fixing this.

	f       = io.open( file, "rb" );
	size    = fileSize( f );

	c_count = 0
	bytes_read = 0
	chunkFailures = 0;
	while true do
		bytes = f:read(chunksize)
		if bytes ~= nil then
			b_sent = 0;
			iterations = 0;
			chunklength = string.len(bytes)
			while (b_sent < chunklength) and (iterations < 100) do
				if Controls.read():start() then break end
				b_sent = client:send(bytes)
				if iterations > 0 then
					System.sleep(50)
				else
					System.sleep(1)
				end
				if iterations == 99 then
					graphicsPrint("X")
					chunkFailures = chunkFailures + 1
				end
				iterations = iterations + 1
				graphicsPrint("|")
			end
		end
		c_count = c_count + 1
		graphicsPrint(".")
		System.sleep(1)
		bytes_read = bytes_read + chunksize
		if bytes_read >= size then break end
		if chunkFailures == 1 then break end
		if Controls.read():start() then break end
	end
	if chunkFailures == 1 then
		graphicsPrint("\nThere was an error sending the file.\n")
	else
		graphicsPrint("\nSent file with " .. c_count .. " chunks of " .. chunksize .. " bytes.\n")
	end
	f:close();

	return size, "200"
end

-- start server socket
while true do
	start_server(80)
end
	

Wlan.term()